open Parser

(* type token *)
type token = Comma | Comment | Lparen | Rparen
  | Keyword of Strng.t
  | Number of Strng.t
  | Ident of Strng.t
(* *)

module PlLexer = struct
(* is_comment : token -> bool *)
let is_comment = function Comment -> true | _ -> false
(* keywords : Strng.t list *)
let keywords =
  Lst.map Strng.of_string ["if";"then";"else";"="]
(* lex : ('a,char)t -> ('a,char)t *)
let lex p =
  p      >>= fun x ->
  spaces >>= fun _ ->
  return x
(* comment : (token,char)t *)
let comment =
  char '#'          >>= fun _ ->
  many(noneof "\n") >>= fun _ ->
  return Comment
(* number : (token,char)t *)
let number =
  ?> (oneof "+-") '+' >>= fun sign ->
  many1 digit         >>= fun ds   ->
  return(Number(sign::ds))
(* ident : (token,char)t *)
let ident =
  many1(noneof "(), \n\r\t") >>= fun cs ->
  if Lst.mem cs keywords then return(Keyword cs)
                         else return(Ident cs)
(* lparen : (token,char)t *)
let lparen = char '(' >>= fun _ -> return Lparen
(* rparen : (token,char)t *)
let rparen = char ')' >>= fun _ -> return Rparen
(* comma : (token,char)t *)
let comma  = char ',' >>= fun _ -> return Comma
(* token : (token,char)t *)
let token = lex(
  comment <|> number <|> ident <|>
  lparen  <|> rparen <|> comma
)
(* tokenize : (token list,char)t *)
let tokenize =
  spaces     >>= fun _  ->
  many token >>= fun ts ->
  eoi        >>= fun _  ->
  return ts
(* *)
end


(* type sym *)
type sym = Strng.t
(* type ast *)
type ast =
  | Const of Strng.t
  | Ite of (ast * ast * ast)
  | Call of (sym * ast list)
  | Def of (sym * sym list * ast * ast)
(* *)

module PlParser = struct
(* PlParser.kwd : string -> (char list,token)t *)
let kwd s  = token(function
  | Keyword k when k = Strng.of_string s -> Some k
  | _ -> None)
(* PlParser.number : (int,token)t *)
let number = token(function
  | Number n -> Some n
  | _ -> None)
(* PlParser.ident : (char list,token)t *)
let ident  = token(function Ident i -> Some i | _ -> None)
(* PlParser.lparen : (unit,token)t *)
let lparen = token(function Lparen -> Some() | _ -> None)
(* PlParser.rparen : (unit,token)t *)
let rparen = token(function Rparen -> Some() | _ -> None)
(* PlParser.comma : (unit,token)t *)
let comma  = token(function Comma -> Some() | _ -> None)
(* *)

(* par : ('a,token)t -> ('a,token)t *)
let par p = between lparen p rparen
(* parm : ('a,token)t -> ('a list,token)t *)
let parm p = between lparen (sep_by comma p) rparen
(* *)

(* const : (ast,token)t *)
let const = number >>= fun n -> return(Const n)
(* def : unit -> (ast,token)t *)
let rec def() =
  ident              >>= fun f  ->
  ?> (parm ident) [] >>= fun xs ->
  kwd "="            >>= fun _  ->
  exp()              >>= fun e1 ->
  exp()              >>= fun e2 ->
  return(Def(f,xs,e1,e2))
(* call : unit -> (ast,token)t *)
and call() =
  ident               >>= fun f  ->
  ?> (parm(exp())) [] >>= fun es ->
  return(Call(f,es))
(* ite : unit -> (ast,token)t *)
and ite() =
  kwd "if"   >>= fun _ ->
  exp()      >>= fun b ->
  kwd "then" >>= fun _ ->
  exp()      >>= fun t ->
  kwd "else" >>= fun _ ->
  exp()      >>= fun e ->
  return(Ite(b,t,e))
(* exp : unit -> (ast,token)t *)
and exp() =
  def() <|> ite() <|> call() <|> const <|> (
    lparen >>= fun _ ->
    exp()  >>= fun e ->
    rparen >>= fun _ ->
    return e
)
(* ast : (ast,token)t *)
let ast =
  exp() >>= fun e ->
  eoi   >>= fun _ ->
  return e
(* *)
end
